package ava.parser

/** Enumeration for all supported character escapes.
  *
  * @param name
  *   The character escape name - how it resolves in source code.
  * @param output
  *   The character emitted by the escape.
  */
sealed abstract class CharEscape(
  val name: Char,
  val output: Char
)

object CharEscape:

  case object NewLine        extends CharEscape('n', '\n')
  case object CarriageReturn extends CharEscape('r', '\r')
  case object Backspace      extends CharEscape('b', '\b')
  case object FormFeed       extends CharEscape('f', '\f')
  case object Tab            extends CharEscape('t', '\t')
  case object BackSlash      extends CharEscape('\\', '\\')
  case object DoubleQuote    extends CharEscape('"', '"')
  case object SingleQuote    extends CharEscape('\'', '\'')

  val All: List[CharEscape] = List(
    NewLine,
    CarriageReturn,
    Backspace,
    FormFeed,
    Tab,
    BackSlash,
    DoubleQuote,
    SingleQuote
  )

  /** Given some escape name, attempt to resolve to a supported [[CharEscape]].
    *
    * @param escapeName
    *   The candidate name.
    * @return
    *   The relevent [[CharEscape]], or `None` if no such escape exists.
    */
  def resolve(escapeName: Char): Option[CharEscape] =
    All.find(_.name == escapeName)

end CharEscape
