package ava.parser

import java.io.BufferedReader
import java.io.FileReader
import java.io.InputStream
import java.io.InputStreamReader
import java.nio.charset.StandardCharsets
import java.nio.file.Path

/** Used to consume characters from the input stream.
  *
  * @param input
  *   Input stream capable of reading _characters_ rather than bytes. This
  *   stream is owned by the character reader.
  */
class CharacterReader(
  private val input: BufferedReader
):
  // Internal constants.
  private val Capacity: Int         = 4096
  private val LookBackCapacity: Int = 16

  // Internal buffers.
  private var currentChar: Char     = 0
  private val lastChars: Ring[Char] = Ring[Char](LookBackCapacity, 0)
  private val buffer: Array[Char]   = Array.fill(Capacity)(0)

  // Internal mutable state.
  private var length: Int          = 0
  private var index: Int           = 0
  private var eof: Boolean         = false
  private var peekedAhead: Boolean = false
  private var lookAhead: Char      = 0

  // Tracking for callers.
  private var charInFile: Long = 0
  private var lineNumber: Long = 1
  private var charInLine: Long = 0

  /** Close the underlying stream.
    */
  def close(): Unit = input.close()

  def currentSourcePosition(): SourcePosition =
    SourcePosition(charInFile, lineNumber, charInLine)

  /** Set EOF and close the underlying stream.
    */
  private def setEof(): Unit =
    eof = true
    close()

  /** Used when the buffer has been fully consumed. Reads as many characters as
    * possible and resets internal state. Detects EOF and sets state
    * accordingly.
    */
  private def fillBuffer(): Unit =
    if eof then ()
    else
      var numberOfCharacters = 0
      if peekedAhead then
        // Account for the case where we were forced to look ahead of the
        // current stream and consume a character. This means we have to set
        // that consumed character as the first element of the buffer.
        buffer(0) = lookAhead
        numberOfCharacters = input.read(buffer, 1, Capacity - 1)
      else numberOfCharacters = input.read(buffer, 0, Capacity)

      // Record the number of characters ACTUALLY consumed by the stream.
      // This is our working buffer size.
      length = numberOfCharacters

      // If no characters could be read, we're done.
      /* EDGE CASE: If we peeked ahead to the last character, we still
       * technically */
      // have a valid buffer of length 1.
      val _ =
        if numberOfCharacters <= 0 && peekedAhead then
          // Edge case: buffer of size 1.
          length = 1
        else if length <= 0 then
          // EOF case: no characters remain.
          length = 0
          setEof()
          input.close()
        else
          // Normal case: we read some number of characters.
          length = numberOfCharacters + (if peekedAhead then 1 else 0)

      // Reset the peeked state -- we already consumed it.
      peekedAhead = false

      // Reset the index position to start iterating through the buffer again.
      index = 0

  /** Observe, but do not consume, the next character in the stream. Note that
    * if the buffer is fully consumed, this function must read the underlying
    * stream but will _not_ refresh the buffer until the next character is
    * consumed.
    *
    * @return
    *   The next character value, or None if the stream is EOF.
    */
  def peek(): Option[Char] =
    if eof then None
    else if index == length then
      // Special case -- we have exhausted the buffer and still want to peek.
      // In this case, we consume exactly one character (if possible) and set
      // the peek state.
      val ch = input.read()
      if ch < 0 then
        eof = true
        None
      else
        peekedAhead = true
        lookAhead = ch.toChar
        Some(lookAhead)
    else Some(buffer(index))

  def consume(): Option[Char] =
    if eof then None
    else if peekedAhead then
      // If we are in the "peeked ahead" state, we know that we were forced to
      // consume the stream to peek. Fill the buffer accordingly.
      fillBuffer()

      // buffer(0) contains what we need, but it's ALSO in lookAhead. Take it,
      // increment the index (to 1), and return the correct value.
      val selectedChar = buffer(0)
      updateCurrentChar(selectedChar)
      index = index + 1
      Some(selectedChar)
    else if index >= length then
      // The buffer has been exhausted. Refill it.
      fillBuffer()
      if eof then None
      else
        // At this point, 'index' should be 0.
        val pos = index
        index = index + 1
        val ch = buffer(pos)
        updateCurrentChar(ch)
        Some(ch)
    else
      // Regular case -- the buffer is not exhausted and contains data. Get the
      // data at the current position and move our pointer.
      val pos = index
      index = index + 1
      val ch = buffer(pos)
      updateCurrentChar(ch)
      Some(ch)

  /** @return
    *   True if all data in the _buffer_ has been consumed, false otherwise.
    */
  def isBufferExhausted(): Boolean =
    index >= length

  /** @return
    *   True if the stream is EOF, false otherwise. An EOF stream does NOT imply
    *   that the buffer is exhausted.
    */
  def isEof(): Boolean =
    eof

  /** Get the _previous_ character. If 0 or 1 characters have been read in
    * total, this will throw an exception due to being uninitialized.
    *
    * @return
    *   The previous character.
    */
  def getPreviousChar(): Char = lastChars.newest()

  /** Get the _current_ character. Initialized to 0.
    *
    * @return
    *   The current character, or 0 if nothing has been consumed.
    */
  def getCurrentChar(): Char = currentChar

  /** Dump the lookback array from newest character to oldest character.
    *
    * @return
    *   The lookback array, ordered from newest to oldest.
    */
  def getLookback(): Array[Char] = lastChars.newestToOldest()

  private def updateCurrentChar(ch: Char): Unit =
    val _ =
      if currentChar != 0 then lastChars.push(currentChar)
      else ()

    // Update tracking information.
    charInFile = charInFile + 1
    if ch == '\n' then
      lineNumber = lineNumber + 1
      charInLine = 0
    else charInLine = charInLine + 1

    // Finally, remap the current character.
    currentChar = ch

object CharacterReader:

  /** Initialize a [[CharacterReader]] for the the given file, using the UTF-8
    * character set.
    *
    * @param path
    *   The path to the file.
    * @return
    *   The new [[CharacterReader]].
    */
  def forPath(path: Path): CharacterReader =
    new CharacterReader(
      new BufferedReader(
        new FileReader(path.toFile(), StandardCharsets.UTF_8)
      )
    )

  /** Initialize a [[CharacterReader]] for the given input stream, using the
    * UTF-8 character set.
    *
    * @param inputStream
    *   The input stream.
    * @return
    *   The new [[CharacterReader]].
    */
  def forInputStream(inputStream: InputStream): CharacterReader =
    new CharacterReader(
      new BufferedReader(
        new InputStreamReader(inputStream, StandardCharsets.UTF_8)
      )
    )

end CharacterReader
