package ava.parser

import scala.reflect.ClassTag

/** Basic ring buffer implementation.
  *
  * @param capacity
  *   Buffer capacity.
  * @param unit
  *   The unit value -- used to populate the initial buffer.
  */
final class Ring[A: ClassTag] private (
  val capacity: Int,
  val unit: A
):
  private var currentSize: Int = 0
  private var oldestIndex: Int = -1
  private var newestIndex: Int = -1
  private val data: Array[A]   = Array.fill(capacity)(unit)

  /** Push new data into this buffer. If the buffer is at capacity, replace the
    * oldest item in the buffer.
    *
    * @param value
    *   The value to push into the buffer.
    */
  def push(value: A): Unit =
    if currentSize <= 0 then
      // Only used the first time data is pushed to the ring.
      data(0) = value
      oldestIndex = 0
      newestIndex = 0
      currentSize = 1
    else if currentSize < capacity then
      // Used when filling up the ring.
      data(currentSize) = value
      newestIndex = currentSize
      currentSize = currentSize + 1
    else
      // We have met the capacity and need to start rolling through the oldest
      // entries, one at a time
      data(oldestIndex) = value
      newestIndex = oldestIndex
      oldestIndex = calculateNextOldest(oldestIndex)

  /** Get the value of the newest item in the buffer. Does not care whether the
    * buffer has been populated.
    *
    * @return
    *   The newest value in the buffer.
    */
  def newest(): A =
    if newestIndex < 0 then
      throw new IllegalStateException(
        "This ring has not been initialized with any data."
      )
    else data(newestIndex)

  /** Extract a copy of this ring, ordered from newest to oldest.
    *
    * @return
    *   Ordered copy of this ring from newest to oldest.
    */
  def newestToOldest(): Array[A] =
    val output      = new Array[A](currentSize)
    var outputIndex = 0
    var ringPointer = newestIndex
    while outputIndex < currentSize do
      output(outputIndex) = data(ringPointer)
      outputIndex = outputIndex + 1
      ringPointer = calculateNextNewest(ringPointer)
    output

  /** Get the current size of the buffer.
    *
    * @return
    *   The current size of the buffer.
    */
  def size(): Int = currentSize

  private def calculateNextOldest(from: Int): Int =
    if from < (currentSize - 1) then from + 1
    else 0

  private def calculateNextNewest(from: Int): Int =
    if from > 0 then from - 1
    else currentSize - 1

object Ring:

  def apply[A: ClassTag](
    capacity: Int,
    unit: A
  ): Ring[A] =
    if capacity > 0 then new Ring[A](capacity, unit)
    else
      throw new IllegalArgumentException(
        "Rings may only be constructed with capacity > 0"
      )

end Ring
