package ava.parser

sealed trait Token

object Token:
  given CanEqual[Token, Token] = CanEqual.derived

  /** Most tokens are generic tokens. They represent some arbitrary grouping of
    * characters that will be refined later.
    *
    * @param value
    *   The token value.
    */
  case class Generic(value: String) extends Token
  given CanEqual[Generic, Generic] = CanEqual.derived

  /** Comments are detected at time of tokenization, and are arbitrary strings.
    *
    * @param value
    *   The comment value.
    */
  case class Comment(value: String) extends Token
  given CanEqual[Comment, Comment] = CanEqual.derived

  /** String literals are captured in total, with escapes resolved.
    *
    * @param value
    *   The resolved string literal value.
    * @param errors
    *   The list of errors that occurred while resolving this string literal, if
    *   any.
    */
  case class StringLiteral(
    value: String,
    errors: List[Tokenizer.Error]
  ) extends Token

  given CanEqual[StringLiteral, StringLiteral] = CanEqual.derived

  /** Character literals are captured in total, with escapes resolved.
    *
    * @param value
    *   The resolved character literal value.
    * @param errors
    *   The list of errors that occurred while resolving this character literal,
    *   if any.
    */
  case class CharacterLiteral(
    value: Char,
    errors: List[Tokenizer.Error]
  ) extends Token

  given CanEqual[CharacterLiteral, CharacterLiteral] = CanEqual.derived

  /** The '(' character.
    */
  case object OpenParen extends Token

  /** The ')' character.
    */
  case object CloseParen extends Token

  /** The '.' character.
    */
  case object Dot extends Token

  /** The ',' character.
    */
  case object Comma extends Token

  /** The ':' character.
    */
  case object Colon extends Token

  /** The '#' character.
    */
  case object Tuple extends Token

  /** Represents end of file (EOF), not in error.
    */
  case object Eof extends Token
end Token
