package ava.parser

import cats.effect.IO
import cats.effect.unsafe.IORuntime
import java.io.ByteArrayInputStream
import java.io.InputStream
import scala.io.Source

class CharacterReaderTests extends munit.FunSuite:
  import CharacterReaderTests.*

  implicit val runtime: IORuntime = cats.effect.unsafe.IORuntime.global

  test("should read a file end to end") {
    val file     = Files.TestSource1
    val expected = loadFileToString(file)
    val reader   = CharacterReader.forInputStream(loadFileToStream(file))

    val output = fs2.Stream
      .repeatEval(IO(reader.consume()))
      .takeWhile(_.isDefined)
      .compile
      .toList
      .map(_.flatten)
      .map(_.mkString)
      .unsafeRunSync()

    assertEquals(output, expected)
    assertEquals(reader.isEof(), true)
    assertEquals(reader.isBufferExhausted(), true)
  }

  test("should properly handle an empty stream") {
    val stream = new ByteArrayInputStream(Array[Byte]())
    val reader = CharacterReader.forInputStream(stream)

    // Verify initial state.
    assertEquals(0, stream.available())
    assertEquals(reader.isBufferExhausted(), true)
    assertEquals(reader.isEof(), false)

    // Trigger recognition of EOF.
    // Case 1 Peek: index == length (initial state)
    assertEquals(reader.peek(), None)
    // Case 2 Peek: eof == true (triggered state)
    assertEquals(reader.peek(), None)
    // Case: eof == true (triggered state)
    assertEquals(reader.consume(), None)
    assertEquals(reader.isBufferExhausted(), true)
    assertEquals(reader.isEof(), true)
    reader.close()
    assertEquals(reader.isBufferExhausted(), true)
    assertEquals(reader.isEof(), true)
  }

  test("should receive the current and previous characters") {
    val stream = new ByteArrayInputStream(Array[Byte]('a', 'b'))
    val reader = CharacterReader.forInputStream(stream)

    // Verify initial state.
    assertEquals(0, reader.getCurrentChar().toInt)
    interceptMessage[IllegalStateException](
      "This ring has not been initialized with any data."
    ) {
      reader.getPreviousChar()
    }

    // Perform this test without peaking, just consume characters.
    assertEquals(reader.consume(), Some('a'))
    assertEquals(reader.isBufferExhausted(), false)
    assertEquals(reader.isEof(), false)
    assertEquals(reader.getCurrentChar(), 'a')
    assertEquals(reader.consume(), Some('b'))
    assertEquals(reader.isBufferExhausted(), true)
    assertEquals(reader.getCurrentChar(), 'b')
    assertEquals(reader.getPreviousChar(), 'a')
    assertEquals(reader.getLookback().toList, List('a'))
    assertEquals(reader.consume(), None)
    assertEquals(reader.isBufferExhausted(), true)
    assertEquals(reader.isEof(), true)
  }

  private def loadFileToString(name: String): String =
    Source.fromResource(name).mkString

  private def loadFileToStream(name: String): InputStream =
    getClass().getClassLoader().getResourceAsStream(name)

object CharacterReaderTests:

  object Files:
    val TestSource1: String = "test-source-1.ava"
