package ava.parser

import cats.effect.IO
import cats.effect.Resource
import cats.effect.unsafe.IORuntime
import java.io.ByteArrayInputStream
import java.io.InputStream
import scala.io.Source

class TokenizerTests extends munit.FunSuite:

  import Tokenizer.Error
  import Tokenizer.State

  implicit val runtime: IORuntime = cats.effect.unsafe.IORuntime.global

  private val EmptyStream: InputStream = new ByteArrayInputStream(Array[Byte]())

  test("should produce the EOF token for an empty stream") {
    run(newCR(EmptyStream)) { tokenizer =>
      IO(assertEquals(tokenizer.next(), Right(Token.Eof)))
    }
  }

  test(
    "should produce the EOF token for a stream of only whitespace characters"
  ) {
    run(newCR(stringStream(" \t\n\r\n   "))) { tokenizer =>
      IO(assertEquals(tokenizer.next(), Right(Token.Eof)))
    }
  }

  test(
    "should produce an error, followed by EOF, if only a backslash chararacter is present"
  ) {
    assertTokens(
      " \\ ",
      Left(Error.BackSlashNotAllowed(pos(2, 1, 2)))
    )
  }

  test("should capture all fixed, single-character tokens") {
    assertTokens(
      "(),:.#",
      Right(Token.OpenParen),
      Right(Token.CloseParen),
      Right(Token.Comma),
      Right(Token.Colon),
      Right(Token.Dot),
      Right(Token.Tuple)
    )
  }

  test("should not recognize comments if preceded by any non-whitespace") {
    assertTokens(
      "( --",
      Right(Token.OpenParen),
      Right(Token.Generic("--"))
    )
  }

  test("should recognize potential comments that are just a dash (space)") {
    assertTokens("- ", Right(Token.Generic("-")))
  }

  test("should recognize potential comments that are just a dash (tab)") {
    assertTokens("-\t", Right(Token.Generic("-")))
  }

  test("should recognize potential comments that are just a dash (newline)") {
    assertTokens("-\n", Right(Token.Generic("-")))
  }

  test("should capture a generic - if a potential comment hits EOF") {
    assertTokens("-", Right(Token.Generic("-")))
  }

  test("should handle basic comments") {
    assertTokens("-- Comment", Right(Token.Comment("Comment")))
  }

  test("should handle comments, if only whitespace precedes them") {
    assertTokens("\t   -- Comment", Right(Token.Comment("Comment")))
  }

  test("should handle successive lines of comments") {
    assertTokens(
      "--c1\n--c2\n--c3",
      Right(Token.Comment("c1")),
      Right(Token.Comment("c2")),
      Right(Token.Comment("c3"))
    )
  }

  test("should ignore comment contents") {
    assertTokens("-- a\\12 3\t4", Right(Token.Comment("a\\12 3\t4")))
  }

  test("should throw an error if EOF is reached inside a string literal") {
    assertTokens(
      "\"",
      Left(
        Error.PrematureEof(
          List(State.InDoubleQuote(pos(1, 1, 1)))
        )
      )
    )
  }

  test("should throw an error if EOF is reached inside a character literal") {
    assertTokens(
      "\'",
      Left(
        Error.PrematureEof(
          List(State.InSingleQuote(pos(1, 1, 1)))
        )
      )
    )
  }

  test("should throw an error if EOF is reached inside a character ESCAPE") {
    assertTokens(
      "\'\\",
      Left(
        Error.PrematureEof(
          List(
            State.InCharEscape(pos(2, 1, 2)),
            State.InSingleQuote(pos(1, 1, 1))
          )
        )
      )
    )
    assertTokens(
      "\"\\",
      Left(
        Error.PrematureEof(
          List(
            State.InCharEscape(pos(2, 1, 2)),
            State.InDoubleQuote(pos(1, 1, 1))
          )
        )
      )
    )
  }

  test("should capture all supported char escapes in a string literal") {
    assertTokens(
      "\"\\n\\r\\b\\f\\t\\\\\\\"\\\'\"",
      Right(Token.StringLiteral("\n\r\b\f\t\\\"\'", Nil))
    )
  }

  test("should capture all supported char escapes in character literals") {
    assertTokens("\'\\n\'", Right(Token.CharacterLiteral('\n', Nil)))
    assertTokens("\'\\r\'", Right(Token.CharacterLiteral('\r', Nil)))
    assertTokens("\'\\b\'", Right(Token.CharacterLiteral('\b', Nil)))
    assertTokens("\'\\f\'", Right(Token.CharacterLiteral('\f', Nil)))
    assertTokens("\'\\t\'", Right(Token.CharacterLiteral('\t', Nil)))
    assertTokens("\'\\\\\'", Right(Token.CharacterLiteral('\\', Nil)))
    assertTokens("\'\\\"\'", Right(Token.CharacterLiteral('"', Nil)))
    assertTokens("\'\\'\'", Right(Token.CharacterLiteral('\'', Nil)))
  }

  test("should handle a basic string literal") {
    assertTokens("\"foo\"", Right(Token.StringLiteral("foo", Nil)))
  }

  test("should handle a string literal with an invalid character escape") {
    // Note: The character escape is simply not emitted, but we are able to
    // successfully close out the string, which allows us to recover and parse
    // more content.
    assertTokens(
      "\"foo\\!bar\"",
      Right(
        Token.StringLiteral(
          "foobar",
          List(Error.InvalidCharEscape(pos(6, 1, 6), '!'))
        )
      )
    )
  }

  test("should reject a string literal with a newline") {
    assertTokens(
      "\"\n",
      Left(
        Error.UnexpectedNewLine(
          sourcePosition = pos(2, 2, 0),
          currentState = State.InDoubleQuote(pos(1, 1, 1))
        )
      )
    )
  }

  test("should reject a character literal with a newline") {
    assertTokens(
      "\'\n",
      Left(
        Error.UnexpectedNewLine(
          sourcePosition = pos(2, 2, 0),
          currentState = State.InSingleQuote(pos(1, 1, 1))
        )
      )
    )
  }

  test("should accept an empty string literal") {
    assertTokens("\"\"", Right(Token.StringLiteral("", Nil)))
  }

  test("should reject an empty character literal") {
    assertTokens(
      "\'\'",
      Left(Error.NoCharactersInLiteral(pos(1, 1, 1), Nil))
    )
  }

  test("should reject an empty character literal with a failed escape") {
    assertTokens(
      "\'\\!\'",
      Left(
        Error.NoCharactersInLiteral(
          sourcePosition = pos(1, 1, 1),
          List(Error.InvalidCharEscape(pos(3, 1, 3), '!'))
        )
      )
    )
  }

  test("should reject a character literal with multiple characters") {
    assertTokens(
      "\'abc\'",
      Left(
        Error.MultipleCharactersInLiteral(
          sourcePosition = pos(1, 1, 1),
          candidate = "abc",
          errors = Nil
        )
      )
    )
  }

  test("should consume a generic token") {
    assertTokens("abcdefg", Right(Token.Generic("abcdefg")))
  }

  test("should consume a generic token for every possible keyword") {
    Keyword.All.foreach { keyword =>
      assertTokens(keyword.value, Right(Token.Generic(keyword.value)))
    }
  }

  test("should consume a generic token for every trivial operator") {
    Operator.Trivial.foreach { op =>
      assertTokens(op.value, Right(Token.Generic(op.value)))
    }
  }

  test("should tokenize a valid file (case 1)") {
    val source = loadFileToString("tokenizer-1.ava")
    println(source)
    assertTokens(
      source,
      Right(Token.Generic("namespace")),
      Right(Token.Generic("unit")),
      Right(Token.Dot),
      Right(Token.Generic("test"))
    )
  }

  test("should tokenize a valid file (case 2)") {
    assertTokens(
      loadFileToString("tokenizer-2.ava"),
      Right(Token.Generic("namespace")),
      Right(Token.Generic("unit")),
      Right(Token.Dot),
      Right(Token.Generic("test")),
      Right(Token.Generic("import")),
      Right(Token.Generic("one")),
      Right(Token.Dot),
      Right(Token.Generic("two")),
      Right(Token.Dot),
      Right(Token.Generic("three")),
      Right(Token.Dot),
      Right(Token.Generic("*")),
      Right(Token.Generic("import")),
      Right(Token.Generic("x")),
      Right(Token.Dot),
      Right(Token.Generic("y")),
      Right(Token.Generic("fn")),
      Right(Token.Generic("foo")),
      Right(Token.Colon),
      Right(Token.Generic("Int")),
      Right(Token.Generic("->")),
      Right(Token.Generic("Int")),
      Right(Token.Generic("λ")),
      Right(Token.Generic("value")),
      Right(Token.Generic("=>")),
      Right(Token.Generic("value")),
      Right(Token.Generic("end")),
      Right(Token.Generic("fn"))
    )
  }

  test("should tokenize a valid file (case 3)") {
    val source = loadFileToString("tokenizer-3.ava")
    println(source)
    assertTokens(
      source,
      Right(Token.Generic("namespace")),
      Right(Token.Generic("unit")),
      Right(Token.Dot),
      Right(Token.Generic("test")),
      Right(
        Token.Comment(
          "- Type class for type constructors which can be mapped over."
        )
      ),
      Right(Token.Generic("given")),
      Right(Token.Generic("F")),
      Right(Token.Generic("*")),
      Right(Token.Generic("class")),
      Right(Token.Generic("Functor")),
      Right(
        Token.Comment("- Transform some wrapped data from one type to another.")
      ),
      Right(Token.Generic("given")),
      Right(Token.Generic("A")),
      Right(Token.Comma),
      Right(Token.Generic("B")),
      Right(Token.Generic("defn")),
      Right(Token.Generic("map")),
      Right(Token.Colon),
      Right(Token.Generic("F")),
      Right(Token.Generic("A")),
      Right(Token.Generic("->")),
      Right(Token.OpenParen),
      Right(Token.Generic("A")),
      Right(Token.Generic("->")),
      Right(Token.Generic("B")),
      Right(Token.CloseParen),
      Right(Token.Generic("->")),
      Right(Token.Generic("F")),
      Right(Token.Generic("B")),
      Right(Token.Generic("end")),
      Right(Token.Generic("class"))
    )
  }

  private def assertTokens(
    source: String,
    expectedOutput: Either[Error, Token]*
  ): Unit =
    run(newCR(stringStream(source))) { tokenizer =>
      consumeAll(tokenizer).map { tokens =>
        assertEquals(tokens, expectedOutput.toList)
      }
    }

  private def consumeAll(tokenizer: Tokenizer): IO[List[Either[Error, Token]]] =
    fs2.Stream
      .repeatEval(IO(tokenizer.next()))
      .takeWhile(_ != Right(Token.Eof))
      .compile
      .toList

  private def pos(
    charInFile: Long,
    lineNumber: Long,
    charInLine: Long
  ): SourcePosition =
    SourcePosition(charInFile, lineNumber, charInLine)

  private def stringStream(contents: String): InputStream =
    new ByteArrayInputStream(contents.getBytes())

  private def newCR(is: InputStream): CharacterReader =
    CharacterReader.forInputStream(is)

  private def run(reader: CharacterReader)(testF: Tokenizer => IO[Unit]): Unit =
    Resource
      .make(
        acquire = IO(new Tokenizer(reader))
      )(release = t => IO(t.close()))
      .use(testF)
      .unsafeRunSync()

  private def loadFileToString(name: String): String =
    Source.fromResource(name).mkString
